/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: DistributedJMXServerManager.java 16:34:49 ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.communication.jmx;

import java.util.HashMap;
import java.util.Map;

import javax.management.MBeanServer;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;

import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.fractal.jmx.agent.AdminAttributes;
import org.objectweb.petals.PetalsException;
import org.objectweb.petals.communication.jmx.rmi.RMIConnector;
import org.objectweb.petals.communication.network.ContainerInformation;
import org.objectweb.petals.communication.network.NetworkService;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * This class is a manager of JMX server. It is used to get a connection to a
 * JMX Server on a distant container.
 *
 * @author ddesjardins - eBMWebsourcing
 * @author chamerling - eBMWebsourcing
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.objectweb.petals.communication.jmx.JMXService.class))
public class JMXServiceImpl implements JMXService {

    /**
     * Logger
     */
    protected LoggingUtil log;

    @Monolog(name = "logger")
    private Logger logger;

    /**
     * AdminAttributes
     */
    @Requires(name = "adminAtt", signature = org.objectweb.fractal.jmx.agent.AdminAttributes.class)
    protected AdminAttributes adminAttributes;

    @Requires(name = "network", signature = org.objectweb.petals.communication.network.NetworkService.class)
    protected NetworkService containerNetwork;

    /**
     * Create a {@link DistributedJMXServer} instance
     *
     * @param containerName
     * @return
     * @throws PetalsException
     */
    public DistributedJMXServer createDistributedJMXServer(String containerName)
        throws PetalsException {
        log.start();
        DistributedJMXServer distributedJMXServer = null;
        try {
            ContainerInformation containerInformation = containerNetwork
                .getContainerInformation(containerName);
            if (containerInformation != null) {
                JMXConnector connector = null;
                Map<String, Object> args = new HashMap<String, Object>();
                String username = containerInformation.getJmxLogin();
                String password = containerInformation.getJmxPassword();
                if (!"".equals(username)) {
                    if (password == null) {
                        password = "";
                    }
                    String[] credentials = new String[] {username, password};
                    args.put("jmx.remote.credentials", credentials);
                }

                String urlStr = RMIConnector
                    .getJmxServiceURL(containerInformation.getHost(),
                            containerInformation.getJmxPort(),
                            containerInformation.getJmxJndiPort());
                JMXServiceURL url = new JMXServiceURL(urlStr);
                connector = JMXConnectorFactory.newJMXConnector(url, args);

                distributedJMXServer = new DistributedJMXServer(connector);
            }
        } catch (Exception e) {
            throw new PetalsException(
                "Error while trying to connect to JMX server", e);
        }
        log.end();
        return distributedJMXServer;
    }

    /**
     * @see DistributedJMXServerFactory#isContainerStarted(String)
     */
    public boolean isContainerStarted(String containerName) {
        log.call();
        boolean result = true;
        try {
            containerNetwork.isContainerStarted(containerName);
        } catch (PetalsException e) {
            result = false;
        }
        return result;
    }

    @LifeCycle(on = LifeCycleType.START)
    public void start() throws IllegalLifeCycleException {
        log = new LoggingUtil(logger);
        log.call();
    }

    @LifeCycle(on = LifeCycleType.STOP)
    public void stop() throws IllegalLifeCycleException {
        log.call();
    }

    /**
     * Get the local jmx server
     *
     * @return the locel jmx server
     */
    public MBeanServer getLocalJMXServer() {
        return adminAttributes.getRawMBeanServer();
    }
}
