/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.communication.jndi;

import org.objectweb.petals.PetalsException;
import org.objectweb.petals.util.LoggingUtil;

/**
 * An abstract for the JNDI agent service. The JNDI agent fractal component MUST
 * be started before all the others.
 *
 * @author Christophe HAMERLING - eBMWebSourcing
 *
 */
public abstract class AbstractJNDIAgentService implements JNDIAgentService {

    /**
     * Status of the JNDI agent
     */
    protected Status status;

    /**
     * The JNDI agent runs in dynamic or static mode ?
     */
    protected boolean dynamic;

    /**
     * The host on which the JNDI server runs (local)
     */
    protected String host;

    /**
     * The port used to access to the JNDI server
     */
    protected int port;

    /**
     * Logging utilities
     */
    protected LoggingUtil log;

    /**
     * Creates a new instance of {@link AbstractJNDIAgentService}
     *
     */
    public AbstractJNDIAgentService() {
    }

    /*
     * (non-Javadoc)
     *
     * @see org.objectweb.petals.communication.jndi.JNDIAgentService#initJNDIAgent()
     */
    public void initJNDIAgent() throws PetalsException {
        if (status != Status.NOT_INITIALIZED) {
            throw new PetalsException(getImplName()
                + " : Can not initialize JNDI Agent with status : "
                + getStatus());
        }

        try {
            doInit();
        } catch (Exception e) {
            throw new PetalsException(getImplName()
                + " : Error while initializing the JNDI server", e);
        }
        status = Status.INITIALIZED;
    }

    /*
     * (non-Javadoc)
     *
     * @see org.objectweb.petals.communication.jndi.JNDIAgentService#startJNDIAgent()
     */
    public void startJNDIAgent() throws PetalsException {
        if (status != Status.INITIALIZED) {
            throw new PetalsException(getImplName()
                + " : Can not start JNDI Agent with status : " + getStatus());
        }

        try {
            doStart();
        } catch (Exception e) {
            throw new PetalsException(getImplName()
                + " : Error while starting the JNDI server", e);
        }
        status = Status.STARTED;
    }

    /*
     * (non-Javadoc)
     *
     * @see org.objectweb.petals.communication.jndi.JNDIAgentService#stopJNDIAgent()
     */
    public void stopJNDIAgent() throws PetalsException {
        if (status != Status.STARTED) {
            throw new PetalsException(getImplName()
                + " : Can not stop JNDI Agent with status : " + getStatus());
        }

        try {
            doStop();
        } catch (Exception e) {
            throw new PetalsException(getImplName()
                + " : Error while stopping the JNDI server", e);
        }
        status = Status.STOPPED;
    }

    /*
     * (non-Javadoc)
     *
     * @see org.objectweb.petals.communication.jndi.JNDIAgentService#getStatus()
     */
    public Status getStatus() {
        return status;
    }

    /**
     * Get the JNDI implementation name
     *
     * @return
     */
    public abstract String getImplName();

    /**
     *
     * @throws Exception
     */
    public abstract void doInit() throws Exception;

    /**
     *
     * @throws Exception
     */
    public abstract void doStart() throws Exception;

    /**
     *
     * @throws Exception
     */
    public abstract void doStop() throws Exception;
}
