/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: SocketServerThread.java 11:55:50 AM ddesjardins $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.communication.jndi.tribe.thread;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;

import javax.naming.Binding;
import javax.naming.NameClassPair;
import javax.naming.NamingException;

import org.objectweb.petals.communication.jndi.tribe.TribeServer;
import org.objectweb.petals.communication.jndi.tribe.msg.request.RegistryRequest;
import org.objectweb.petals.communication.jndi.tribe.msg.response.AckResponse;
import org.objectweb.petals.communication.jndi.tribe.msg.response.CreateSubcontextResponse;
import org.objectweb.petals.communication.jndi.tribe.msg.response.ExceptionResponse;
import org.objectweb.petals.communication.jndi.tribe.msg.response.ListBindingsResponse;
import org.objectweb.petals.communication.jndi.tribe.msg.response.ListResponse;
import org.objectweb.petals.communication.jndi.tribe.msg.response.LookupLinkResponse;
import org.objectweb.petals.communication.jndi.tribe.msg.response.LookupResponse;
import org.objectweb.petals.communication.jndi.tribe.msg.response.RegistryResponse;
import org.objectweb.petals.util.LoggingUtil;

/**
 * This class is the socket server for the distributed JNDI directory. It
 * receives requests from the naming context implementation client.
 *
 * @author ddesjardins - eBMWebsourcing
 */
public class SocketServerThread extends Thread {

    /**
     * Logger
     */
    protected LoggingUtil log;

    /**
     * Boolean used to control the lifecycle of the socket server
     */
    protected boolean run = true;

    /**
     * Start time of this server
     */
    protected long sender;

    /**
     * JNDI server
     */
    protected TribeServer server;

    /**
     * The socket used for incoming requests
     */
    protected ServerSocket serverSocket;

    /**
     * The server is started ?
     */
    private boolean started = false;

    /**
     * Creates a new instance of {@link SocketServerThread}
     *
     * @param server
     */
    public SocketServerThread(TribeServer server) {
        super("SocketServerThread" + server.getStartTime());
        this.server = server;
        this.log = server.getLog();
        this.sender = server.getStartTime();
    }

    @SuppressWarnings("unchecked")
    public void run() {
        log.call();
        try {
            serverSocket = new ServerSocket();
            serverSocket.bind(new InetSocketAddress(server.getHost(), server
                    .getPort()));

            log.info("TRIBe JNDI server started on " + server.getHost() + ":"
                    + server.getPort());

            RegistryRequest request = null;
            RegistryResponse response = null;
            ObjectInputStream objectInputStream = null;
            ObjectOutputStream objectOutputStream = null;

            Socket socket = null;
            started = true;
            synchronized(server) {
                server.notify();
            }
            while (run) {
                socket = serverSocket.accept();

                // wait for request
                try {
                    objectInputStream = new ObjectInputStream(socket.getInputStream());
                    request = ((RegistryRequest)(objectInputStream.readObject()));
                } catch (IOException e) {
                    // we receive some crappy data
                    request = null;
                } catch (Exception e) {
                    if ((objectInputStream.available()) != 0) {
                        log.warning("Unable to read JNDI Request Stream");
                    }
                    request = null;   
                }

                if (request != null) {
                    try {

                        // process the request
                        Object object = server.processRequest(request);

                        // send the response
                        switch (request.getType()) {
                            /*
                             * BIND
                             */
                            case bind:
                                response = new AckResponse(sender, 0);
                                request.setSender(sender);
                                break;
                            /*
                             * CREATE SUBCONTEXT
                             */
                            case createSubcontext:
                                response = new CreateSubcontextResponse(
                                        (String) object, sender, 0);
                                request.setSender(sender);
                                break;
                            /*
                             * DESTROY SUBCONTEXT
                             */
                            case destroySubcontext:
                                response = new AckResponse(sender, 0);
                                request.setSender(sender);
                                break;
                            /*
                             * LIST
                             */
                            case list:
                                response = new ListResponse(
                                        (NameClassPair[]) object, sender, 0);
                                break;
                            /*
                             * LIST BINDINGS
                             */
                            case listBindings:
                                response = new ListBindingsResponse(
                                        (Binding[]) object, sender, 0);
                                break;
                            /*
                             * LOOKUP
                             */
                            case lookup:
                                response = new LookupResponse(object, sender,
                                        request.getSender());
                                break;
                            /*
                             * LOOKUP LINK
                             */
                            case lookupLink:
                                response = new LookupLinkResponse(object,
                                        sender, request.getSender());
                                break;
                            /*
                             * PING
                             */
                            case ping:
                                response = new AckResponse(sender, request
                                        .getSender());
                                break;
                            /*
                             * REBIND
                             */
                            case rebind:
                                response = new AckResponse(sender, 0);
                                request.setSender(sender);
                                break;
                            /*
                             * RENAME
                             */
                            case rename:
                                response = new AckResponse(sender, 0);
                                request.setSender(sender);
                                break;
                            /*
                             * UNBIND
                             */
                            case unbind:
                                response = new AckResponse(sender, 0);
                                request.setSender(sender);
                                break;
                            default:
                                break;
                        }

                        if (request.getSender() != 0) {
                            // Send the request to the other registries if the
                            // sender is not null
                            try {
                                server.getChannel().send(request);
                            } catch (Exception e) {
                                log
                                        .error(
                                                "Error while send the request to the other registries",
                                                e);
                            }
                        }
                    } catch (NamingException e) {
                        log.error(e.getMessage(), e);
                        response = new ExceptionResponse(e, sender, request
                                .getSender());
                    }
                    // write response to socket
                    objectOutputStream = new ObjectOutputStream(socket
                            .getOutputStream());
                    objectOutputStream.writeObject(response);
                    objectOutputStream.flush();
                    objectOutputStream.close();
                }
            }
        } catch (SocketException e) {
            if (!"Socket closed".equals(e.getMessage())) {
                log.error(e.getMessage(), e);
            }
        } catch (IOException e) {
            log.error("Error while starting socket server", e);
            started = false;
            synchronized(this) {
                this.notify();
            }
            throw new RuntimeException("Error while starting socket server", e);
        }
        started = false;
    }

    /**
     * Shutdown the server socket
     *
     */
    public void shutdown() {
        log.start();
        run = false;
        // Close the server socket
        try {
            serverSocket.close();
            this.interrupt();
        } catch (Exception e) {
            log.error("Problem while closing server socket " + e);
        }
        started = false;
        log.end();
    }

    /**
     *
     * @return
     */
    public boolean isStarted() {
        return started;
    }

}
