/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.jbi.management.task;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.concurrent.atomic.AtomicLong;

import org.apache.commons.io.FileUtils;
import org.objectweb.petals.PetalsException;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.ZipUtil;

/**
 * This task extracts entity (component, shared lib, service assembly) archive
 * into the work directory <br/> Inputs : archive {@link URI} <br/> Outputs :
 * {@link File} pointing on unzipped archive <br/>
 *
 * @author ofabre - EBM Websourcing
 *
 */
public class ExtractTask extends AbstractFileManipulationTask {

    /**
     * logger wrapper
     */
    protected LoggingUtil log;

    /**
     * DateFormat used for repository folders
     */
    private final static SimpleDateFormat FOLDER_DATE_FORMAT = new SimpleDateFormat("MM-dd-yyyy-hhmmss");

    /**
     * Unique extraction identifier, it is used in temporary path name to prevent
     * concurrent extraction.
     */
    private final static AtomicLong EXTRACTION_ID = new AtomicLong(0L);

    public ExtractTask(LoggingUtil log) {
        super();
        this.log = log;
    }

    @SuppressWarnings("unchecked")
    public void execute(HashMap context) throws Exception {

        URI archiveURI = (URI) context.get(TaskContextConstants.ARCHIVE_URI);

        /*
         * Explode Comp
         */
        File unzipRoot = explodeEntityPackage(archiveURI);

        /*
         * Fill context
         */
        context.put(TaskContextConstants.UNZIP_ROOT, unzipRoot);

    }

    /*
     * SimpleDateFormat must be synchronized.
     */
    private static synchronized String getCurrentDateAsString() {
        
        return ExtractTask.FOLDER_DATE_FORMAT.format(new Date(System
                .currentTimeMillis()));
    }

    private File explodeEntityPackage(URI archiveURI) throws PetalsException {
        File zipFile = new File(archiveURI);
        File destFile = new File(getWorkDirectory(), zipFile.getName() + "-"
                + ExtractTask.getCurrentDateAsString() + "-"
                + ExtractTask.EXTRACTION_ID.getAndIncrement());
        try {
            if (!destFile.mkdirs()) {
                throw new PetalsException(
                        "Failed to create temporary work dir: "
                                + destFile.getAbsolutePath());
            }
            this.log.debug("Extracting archive " + zipFile + " to " + destFile);
            ZipUtil.unzipArchive(zipFile, destFile);
            this.log.debug("Archive extracted.");
        } catch (IOException e) {
            try {
                FileUtils.forceDelete(destFile);
            } catch (IOException e1) {
                // Nothing to do, just log
                log.warning("Residual incomplete component folder");
            }
            throw new PetalsException(e);
        }
        return destFile;
    }

    @SuppressWarnings("unchecked")
    public void undo(HashMap context) {
        File tempDir = (File) context.get(TaskContextConstants.UNZIP_ROOT);
        try {
            if (tempDir.exists()) {
                FileUtils.forceDelete(tempDir);
            }
        } catch (IOException e) {
            ;
            log.error("Failed to revert a ExtractTask: " + e.getMessage());
        }
    }

}
