/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): pierre.garcia@inrialpes.fr
 *
 */

package org.objectweb.petals.jbi.messaging.transport;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.ConcurrentModificationException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;

import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.Cardinality;
import org.objectweb.fractal.fraclet.annotation.annotations.type.Contingency;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.petals.communication.jndi.JNDIEndpointCache;
import org.objectweb.petals.communication.network.ContainerInformation;
import org.objectweb.petals.communication.network.NetworkService;
import org.objectweb.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.objectweb.petals.jbi.messaging.registry.EndpointCacheService;
import org.objectweb.petals.jbi.messaging.servicedesc.AbstractEndpoint;
import org.objectweb.petals.jbi.messaging.transport.util.TransporterUtil;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.SystemUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * This class represents a dispatcher used to transmit messages to the right
 * transporter. When a message arrives, it is sent the message to the
 * appropriate transportProtocol.<br>
 * <br>
 * The transportProtocol to use can be specified by two means. By increasing
 * order, here are the means :
 * <ol>
 * <li>To set the <code>transporter</code> in the <em>server.properties</em>
 * configuration file</li>
 * <li>To set the {@link Transporter.PROPERTY_PROTOCOLS_TO_TRY} in the {@link
 * MessageExchangeImpl message exchange}</li>
 * </ol>
 * In both cases, this property can be a single value or a list of value
 * separated by comas <br>
 * <br>
 * 
 * @author pgarcia
 */
@FractalComponent
@Provides(interfaces = {
		@Interface(name = "service", signature = org.objectweb.petals.jbi.messaging.transport.Transporter.class),
		@Interface(name = "dispatcher", signature = org.objectweb.petals.jbi.messaging.transport.TransportProtocolListener.class) })
public class DispatcherImpl implements Transporter, TransportProtocolListener {

	// ------------------------------------------------------------------------
	// --
	// Constants
	// ------------------------------------------------------------------------
	// --

	/**
	 * Interface name (collection) of the protocols available
	 */
	public static final String DELEGATED_ITF_NAME = "delegated-service";

	static final String FAST_LOCAL_POLICY = "fast_local";

	static final String FAST_REMOTE_POLICY = "fast_remote_local";

	private static Map<String, List<String>> PROTOCOL_POLICIES_MAP = new HashMap<String, List<String>>();

	static {

		// Reliability transporters
		List<String> reliabilityLinkedList = Collections
				.synchronizedList(new ArrayList<String>(1));
		reliabilityLinkedList.add(DispatcherImpl.DELEGATED_ITF_NAME + "-"
				+ Transporter.RELIABLE_POLICY);
		PROTOCOL_POLICIES_MAP.put(Transporter.RELIABLE_POLICY,
				reliabilityLinkedList);

		// Fast local transporters
		List<String> fastLocalList = Collections
				.synchronizedList(new ArrayList<String>(1));
		fastLocalList.add(DispatcherImpl.DELEGATED_ITF_NAME + "-"
				+ Transporter.SCOPE_LOCAL);
		PROTOCOL_POLICIES_MAP.put(FAST_LOCAL_POLICY, fastLocalList);

		// Fast remote transporters
		List<String> fastRemoteList = Collections
				.synchronizedList(new ArrayList<String>(1));
		fastRemoteList.add(DispatcherImpl.DELEGATED_ITF_NAME + "-"
				+ Transporter.FAST_POLICY);
		PROTOCOL_POLICIES_MAP.put(FAST_REMOTE_POLICY, fastRemoteList);
	}

	// ------------------------------------------------------------------------
	// --
	// Loggers
	// ------------------------------------------------------------------------
	// --

	/**
	 * Logger
	 */
	@Monolog(name = "logger")
	protected Logger logger;

	/**
	 * Logger Wrapper
	 */
	protected LoggingUtil log;

	// ------------------------------------------------------------------------
	// --
	// Fields
	// ------------------------------------------------------------------------
	// --

	/**
	 * The local container information.
	 */
	protected ContainerInformation containerInformation;

	/**
	 * Map which will be modified by transporters when a normal message is
	 * availabe for a given component
	 */
	protected Map<String, BlockingQueue<TransportProtocol>> messageAvailable;

	/**
	 * Map of MsgExchange that are blocked during a synchronous send
	 */
	protected Map<String, MessageExchangeImpl> waitingExchanges;

	/**
	 * list of default transport protocols to use when a message exchange does
	 * not specified the transport protocols to use.
	 * 
	 * values contained in this list MUST match an existing transport protocol.
	 */
	protected List<String> defaultProtocolsList;

	/**
	 * The list of the threads which are blocking on a poll
	 */
	private Vector<Thread> threadsList;

	/**
	 * Component started flag
	 */
	private boolean componentStarted = false;

	// ------------------------------------------------------------------------
	// --
	// Fractal Client Interfaces
	// ------------------------------------------------------------------------
	// --

	/**
	 * The container Network service
	 */
	@Requires(name = "network", signature = org.objectweb.petals.communication.network.NetworkService.class)
	private NetworkService network;

	/**
	 * Services to which the transport is delegated
	 */
	@Requires(name = DELEGATED_ITF_NAME, signature = TransportProtocol.class, cardinality = Cardinality.COLLECTION)
	protected Map<String, Object> protocols = new HashMap<String, Object>();

	/**
     * Endpoints cache fractal component
     */
	// TODO: This binding is defined as optional because if it is defined mandatory, PEtALS does not start
	// I have no explaination why PEtALS doesn't start if the binding is mandatory.
    @Requires(name = "endpoint-cache", contingency=Contingency.OPTIONAL, signature = EndpointCacheService.class)
    protected EndpointCacheService endpointCache;

    // ------------------------------------------------------------------------
	// --
	// Fractal Lifecycle methods
	// ------------------------------------------------------------------------
	// --
	/**
	 * Init variables.
	 * 
	 * @throws Exception
	 */
	@LifeCycle(on = LifeCycleType.START)
	protected void start() throws Exception {
		log = new LoggingUtil(logger);
		log.start();
		messageAvailable = new ConcurrentHashMap<String, BlockingQueue<TransportProtocol>>(
				10);
		waitingExchanges = new ConcurrentHashMap<String, MessageExchangeImpl>(
				50);
		threadsList = new Vector<Thread>();
		componentStarted = true;
		containerInformation = network.getLocalContainerInformation();
		log.end();
	}

	/**
	 * Delegate the call of {@link org.objectweb.petals.PetalsLifeCycle#stop()
	 * stop()} to transporters.
	 */
	@LifeCycle(on = LifeCycleType.STOP)
	public void stop() throws Exception {
		log.call();
	}

	// ------------------------------------------------------------------------
	// --
	// Transporter Interface Implementation
	// ------------------------------------------------------------------------
	// --

	/**
	 * Add the destination to all available {@link TransportProtocol}
	 * 
	 * @throws TransportException
	 * 
	 * @see TransportProtocol#addDestination(String)
	 */
	public void addDestination(String componentName) throws TransportException {
		log.start();
		// create an entry in message available
		messageAvailable.put(componentName,
				new LinkedBlockingQueue<TransportProtocol>());
		// add the destination to each transporter
		Iterator<String> it = protocols.keySet().iterator();
		boolean exceptionRaised = false;
		TransportProtocol trProtocol;
		String trProtocolName;
		while (it.hasNext()) {
			trProtocolName = it.next();
			trProtocol = (TransportProtocol) protocols.get(trProtocolName);
			try {
				trProtocol.addDestination(componentName);
			} catch (final TransportException e) {
				exceptionRaised = true;
				log.error("Cannot add the destination '" + componentName
						+ "' to the transport protocol " + trProtocolName, e);
			}
		}
		if (exceptionRaised)
			throw new TransportException(
					"An error occured while adding the destination "
							+ componentName);
		log.end();
	}

	/**
	 * Remove destination from all {@link TransportProtocol}
	 * 
	 * @see TransportProtocol#removeDestination(String)
	 */
	public void removeDestination(final String componentName)
			throws TransportException {
		log.start();
		// remove the destination from the messageAvailable
		messageAvailable.remove(componentName);

		// TODO manage remaining messages in the waitingExchange queue
		// and transporter name in the messageAvailable map.

		// remove the destination from the transport protocols
		Iterator<String> it = protocols.keySet().iterator();
		boolean exceptionRaised = false;
		TransportProtocol trProtocol;
		String trProtocolName;
		while (it.hasNext()) {
			trProtocolName = it.next();
			trProtocol = (TransportProtocol) protocols.get(trProtocolName);
			try {
				trProtocol.removeDestination(componentName);
			} catch (TransportException e) {
				exceptionRaised = true;
				log.error("Cannot remove the destination '" + componentName
						+ "' from the transport protocol " + trProtocolName, e);
			}
		}
		if (exceptionRaised)
			throw new TransportException(
					"An error occured while removing the destination '"
							+ componentName + "'");
		log.end();
	}

	/**
	 * @see Transporter#sendSync(MessageExchangeImpl, String, String, long)
	 */
	public MessageExchangeImpl sendSync(MessageExchangeImpl exchange,
			final String componentName, final String containerName,
			final long timeToLive) throws TransportException {
		log.start();

		MessageExchangeImpl responseExchange = null;
		long startTime = 0;

		// clean unused messages
		exchange.cleanMessages();

		// if the dispatcher is stopped, interrupt the current thread
		if (!componentStarted) {
			Thread.currentThread().interrupt();
			throw new TransportException("The Transporter is stopped",
					new InterruptedException());
		}

		log.debug("Dispatching the transfer of the message '"
				+ exchange.getExchangeId() + "' to component '" + componentName
				+ "' on container '" + containerName + "'");

		// set the Properties to mark the message as a synchronous send
		try {
			TransporterUtil.setSendSyncProperties(exchange);
		} catch (MessagingException e) {
			throw new TransportException(e);
		}

		// mark the stat time to handle the timeout
		if (timeToLive > 0) {
			startTime = System.currentTimeMillis();
		}

		List<String> listProtocols = getTransportProtocolNames(exchange,
				containerName);

		MultiTransportException multiTransportException = null;
		for (String protocolName : listProtocols) {
			log.debug("Try to send via the protocol " + protocolName
					+ " with a timeout of " + timeToLive + " milliseconds");
			TransportProtocol protocol = null;

			try {
				protocol = getProtocol(protocolName);
			} catch (TransportException e) {
				log.error(e.getMessage());
				continue;
			}

			if (timeToLive > 0) {
				// If the timeout is reached, we return null
				if (System.currentTimeMillis() > (startTime + timeToLive)) {
					log
							.warning("Timeout reached for synchronous send of the exchange: "
									+ exchange.getExchangeId());
					log.end();
					return null;
				}
			}

			try {
				responseExchange = sendSyncViaProtocol(exchange, componentName,
						containerName, timeToLive, protocol);
				
				// If the targetted endpoint has been undeployed, the send is
				// succeeded, but an error can be set in the message exchange.
				// We must check this error to invalidate caches.
                // Note: It is unneeded to check on a timeout. It is very
                // unlikely to have a timeout, otherwise the system is very
                // loaded.
				if (responseExchange!= null && responseExchange.getError() != null) {
					// An error has bee set, we invalidate caches
					JNDIEndpointCache.getInstance().cleanCache();
					this.endpointCache.clear();
				}
				
			} catch (TransportException e) {
				JNDIEndpointCache.getInstance().cleanCache();
				this.endpointCache.clear();
				log.warning("The synchronous send with the protocol "
						+ protocolName + " threw a "
						+ TransportException.class.getName(), e);
				if (multiTransportException == null) {
					multiTransportException = new MultiTransportException(
							"The synchronous send failed for multiple reasons. "
									+ "See PEtALS logs or StackTrace for more information",
							e);
				} else {
					multiTransportException.addCount();
				}
				continue;
			}
			break;
		}

		if (responseExchange == null) {
			if (multiTransportException != null
					&& listProtocols.size() == multiTransportException
							.getExceptionCount()) {
				throw multiTransportException;
			} else {
				log.error("Failed to send synchronously the exchange: "
						+ exchange.getExchangeId() + ". Timeout occured");
			}
		}

		log.end();
		return responseExchange;
	}

	/**
	 * Send the message via the specified TransportProtocol(s). the property
	 * {@link Transporter#PROPERTY_TRANSPORT_QOS} can be set in the
	 * <code>messageExchange</code> to specify the transporter to use, otherwise
	 * the {@link Transporter#DEFAULT_POLICY} is used Currently available
	 * transporters are :
	 * <ul>
	 * <li>joram</li>
	 * <li>dream</li>
	 * <li>lcoal</li>
	 * </ul>
	 * 
	 * @see org.objectweb.petals.jbi.messaging.transport.Transporter#send(org.
	 * 	objectweb.petals.jbi.messaging.exchange.MessageExchangeImpl,
	 * 	java.lang.String, java.lang.String, long)
	 */
	public void send(MessageExchangeImpl exchange, String componentName,
			String containerName, long timeToLive) throws TransportException {
		log.start();
		boolean isSent = false;

		// clean unused messages
		exchange.cleanMessages();

		// if the dispatcher is stopped, interrupt the current thread
		if (!componentStarted) {
			Thread.currentThread().interrupt();
			throw new TransportException("The Transporter is stopped",
					new InterruptedException());
		}

		log.debug("Dispatching the transfer of the message '"
				+ exchange.getExchangeId() + "' to component '" + componentName
				+ "' on container '" + containerName + "'");
		List<String> listProtocols = getTransportProtocolNames(exchange,
				containerName);

		TransportProtocol protocol = null;
		for (String protocolName : listProtocols) {
			protocol = getProtocol(protocolName);
			log.debug("Try to send with the protocol " + protocolName
					+ " and a timeout of " + timeToLive + " milliseconds");
			try {
				protocol.send(exchange, componentName, containerName,
						timeToLive);
			} catch (TransportException e) {
				log.warning("The asynchronous send with the protocol "
						+ protocolName + " failed", e);
				continue;
			}
			isSent = true;
			break;
		}

		if (isSent == false) {
			if (SystemUtil.isCacheActivated()) {
				JNDIEndpointCache.getInstance().cleanCache();
				this.endpointCache.clear();
			}

			throw new TransportException(
					"Failed to send asynchronously the exchange: "
							+ exchange.getExchangeId());
		}

		log.end();
	}

	/**
	 * @see
	 * 	org.objectweb.petals.jbi.messaging.transport.Transporter#receive(java
	 * 	.lang.String, long)
	 */
	public MessageExchangeImpl receive(String componentName, long timeoutMS)
			throws TransportException {
		log.start();

		// if the dispatcher is stopped, interrupt the current thread
		if (!componentStarted) {
			Thread.currentThread().interrupt();
			throw new TransportException("The Transporter is stopped",
					new InterruptedException());
		}

		log.debug("Poll a message for the component '" + componentName
				+ "' with a timeout equals to " + timeoutMS);

		BlockingQueue<TransportProtocol> matchingQueue = messageAvailable
				.get(componentName);
		if (matchingQueue == null)
			throw new TransportException("The component '" + componentName
					+ "' destination doesn't exists");

		TransportProtocol transportProtocol;

		if (timeoutMS == 0) {
			// non-blocking call
			transportProtocol = matchingQueue.poll();
		} else {
			threadsList.add(Thread.currentThread());
			try {
				if (timeoutMS > 0) {
					// blocking call with timeout
					transportProtocol = matchingQueue.poll(timeoutMS,
							TimeUnit.MILLISECONDS);
				} else {
					// blocking call without timeout
					transportProtocol = matchingQueue.take();
				}
			} catch (InterruptedException e) {
				Thread.currentThread().interrupt();
				throw new TransportException(
						"The blocking poll for the component '" + componentName
								+ "' is interrupted", e);
			} finally {
				threadsList.remove(Thread.currentThread());
			}
		}

		if (transportProtocol == null) {
			return null;
		}

		final MessageExchangeImpl msgExchange = transportProtocol
				.receive(componentName);
		
		// If the targetted endpoint has been undeployed, the send is
		// succeeded, but an error can be set in the message exchange.
		//  We must check this error to invalidate caches.
		if (msgExchange.getError() != null) {
			// An error has be set, we invalidate caches
			JNDIEndpointCache.getInstance().cleanCache();
			this.endpointCache.clear();
		}

		log.end();

		return msgExchange;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.objectweb.petals.jbi.messaging.transport.Transporter#prepareStop()
	 */
	public void prepareStop() {
		log.start();

		boolean redo = true;
		componentStarted = false;

		// Wait that all the synchronous invocation finished or the timeout is
		// reached
		int lapsTime = 0;
		try {
			while (lapsTime < Transporter.STOP_TIMEOUT) {
				lapsTime += 1000;

				if (waitingExchanges.size() > 0) {
					// sleep for 1 second
					Thread.sleep(1000);
				} else {
					break;
				}
			}
		} catch (InterruptedException e) {
			// do nothing
		}

		if (lapsTime >= STOP_TIMEOUT) {
			log
					.warning("The timeout for stopping the Transporter has been reached."
							+ " Some synchronous invocation are lost.");
		}

		/*
		 * Awake all the pending 'SYNC_SEND' issuer threads to abort the
		 * invocation
		 */
		while (redo) {
			try {
				Collection<MessageExchangeImpl> values = waitingExchanges
						.values();
				for (MessageExchangeImpl messageExchangeImpl : values) {
					synchronized (messageExchangeImpl) {
						messageExchangeImpl.notify();
					}
				}
				redo = false;
			} catch (ConcurrentModificationException e) {
				// Try again
			}
		}

		redo = true;
		/*
		 * Interrupt all the pending receive threads.
		 */
		while (redo) {
			try {
				for (Thread thread : threadsList) {
					thread.interrupt();
				}
				redo = false;
			} catch (ConcurrentModificationException e) {
				// Try again
			}
		}

		log.end();
	}

	// ------------------------------------------------------------------------
	// --
	// PetalsLifeCycle Interface Implementation
	// ------------------------------------------------------------------------
	// --

	/**
	 * Delegate the call of {@link
	 * org.objectweb.petals.PetalsLifeCycle#shutdown() shutdown} to
	 * transporters.
	 * 
	 * @throws Exception
	 */
	public void shutdown() throws Exception {
		log.start();

		boolean exceptionRaised = false;
		for (String protocolName : protocols.keySet()) {
			try {
				((TransportProtocol) protocols.get(protocolName)).shutdown();
			} catch (Exception e) {
				exceptionRaised = true;
				log.error("Cannot remove the transport protocol "
						+ protocolName);
			}
		}
		if (exceptionRaised)
			throw new Exception(
					"An error occured during the shutting down of the dispatcher");

		log.end();
	}

	// ------------------------------------------------------------------------
	// --
	// TransportProtocolListener interface implementation
	// ------------------------------------------------------------------------
	// --

	/**
	 * @see
	 * 	org.objectweb.petals.jbi.messaging.transport.TransportProtocolListener
	 * 	#messageAvailable(java.lang.String, java.lang.String)
	 */
	public void messageAvailable(TransportProtocol transportProtocol,
			String componentName) {
		log.start();

		BlockingQueue<TransportProtocol> matchingQueue = messageAvailable
				.get(componentName);
		if (matchingQueue == null) {
			log.error("The blocking queue for the component " + componentName
					+ " must not be null");
		} else {
			try {
				matchingQueue.put(transportProtocol);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}

		log.end();
	}

	public void syncMessageReceived(MessageExchangeImpl exchange) {
		log.call();

		if (!onMessageExchange(exchange)) {
			sendBackErrorExchange(exchange);
		}
	}

	/**
	 * Receive a response message from a synchronous send.
	 * 
	 * @param exchange
	 * 		The received message exchange
	 * @return <code>true</code> if the response has been correlated with the
	 * 	issuer, <code>false</code> otherwise
	 */
	public boolean onMessageExchange(MessageExchangeImpl exchange) {
		log.start();

		boolean isMatched = false;

		// update the message properties
		TransporterUtil.updateSyncProperties(exchange);

		// push the response message exchange
		if (pushSyncResponse(exchange)) {
			isMatched = true;
		} else {
			// set an error ack to the exchange
			exchange
					.setError(new MessagingException(
							"Reject a message Exchange response from a previous synchronious send"));
			log
					.warning("Reject a message Exchange response from a previous synchronious send. ExchangeId: "
							+ exchange.getExchangeId());
		}

		log.end();
		return isMatched;
	}

	/**
	 * Send synchronously the message exchange via the protocol given as
	 * parameter.
	 * 
	 * @param exchange
	 * @param componentName
	 * @param containerName
	 * @param timeToLive
	 * @param protocol
	 * @return The message exchange containing the reply or <code>null</code> if a timeout occurs.
	 * @throws TransportException
	 */
	private MessageExchangeImpl sendSyncViaProtocol(
			MessageExchangeImpl exchange, String componentName,
			String containerName, long timeToLive, TransportProtocol protocol)
			throws TransportException {
		log.start();

		String exchangeId = TransporterUtil.waitingExchangeKey(exchange);

		// Reference this exchange in the Map of waitingExchanges.
		// It will be used to wake up the exchange when the response
		// is arrived
		waitingExchanges.put(exchangeId, exchange);

		log
				.debug("The message will be sent asynchronously via the given protocol");

		try {
			synchronized (exchange) {
				// Send the message
				protocol.send(exchange, componentName, containerName,
						timeToLive);
				// wait for response or a timeout reached
				exchange.wait(timeToLive);
			}
		} catch (InterruptedException e) {
			waitingExchanges.remove(exchangeId);
			throw new TransportException(
					"Failed to receive the response message of the synchronious send",
					e);
		} catch (TransportException e) {
			waitingExchanges.remove(exchangeId);
			throw e;
		}

		// Now, the response has been sent (wait unblocked),
		// or timeout occured.

		// get the exchange from the map
		MessageExchangeImpl responseExchange = waitingExchanges
				.remove(exchangeId);

		// If the exchange is not the same than the sent exchange,
		// it means that a response message has been received.

		// If it is the same but synchronous property is not set,
		// it means that it is a response from the Local transporter.

		// Otherwise, it means that the timeout has been reached;
		// the response exchange is set to null
		if (responseExchange == exchange
				&& TransporterUtil.getSyncMode(responseExchange, false)) {
			responseExchange = null;
		}

		log.end();
		return responseExchange;
	}

	/**
	 * Send back to the issuer of the response an error ack message
	 * 
	 * @param exchange
	 * 		the message to send back
	 */
	private void sendBackErrorExchange(MessageExchangeImpl exchange) {
		log.start();

		// set the timeout arbitrary to 10 seconds
		try {
			AbstractEndpoint endpoint = null;
			if (exchange.getRole().equals(MessageExchange.Role.PROVIDER)) {
				exchange.setRole(MessageExchange.Role.CONSUMER);
				endpoint = exchange.getConsumerEndpoint();
			} else {
				exchange.setRole(MessageExchange.Role.PROVIDER);
				endpoint = (AbstractEndpoint) exchange.getEndpoint();
			}
			send(exchange, endpoint.getComponentName(), endpoint
					.getContainerName(), 10000L);
		} catch (TransportException e) {
			// do nothing, the message is lost.
		}

		log.end();
	}

	/**
	 * The message is a response of a sendSync() method, the matching sendSync()
	 * must be unblocked with the attached response.
	 * 
	 * @param msg
	 * 		the message received by the <code>Transporter</code>
	 * @throws JBIException
	 * 		The channel has been closed by the component. The message can not be
	 * 		accepted
	 */
	private boolean pushSyncResponse(MessageExchangeImpl responseExchange) {
		log.start();

		boolean isMessageMatched = false;

		String exchangeId = TransporterUtil
				.waitingExchangeKey(responseExchange);

		// Check if this ME is synchronized to a blocked synchronous send.
		// If so, we unblock the sendSync() method, otherwise,
		// we return the message to the local topic, to be consumed by the next
		// accept().
		MessageExchangeImpl sentExchange = waitingExchanges.get(exchangeId);

		if (sentExchange != null) {
			waitingExchanges.put(exchangeId, responseExchange);
			isMessageMatched = true;
			// we notify the exchange issuer that its response is received.
			synchronized (sentExchange) {
				sentExchange.notify();
			}
		}

		log.end();

		return isMessageMatched;
	}

	// ------------------------------------------------------------------------
	// --
	// Utility methods
	// ------------------------------------------------------------------------
	// --

	/**
	 * Get the list of Transport Protocol names to use according to the given
	 * exchange.
	 * 
	 * @param exchange
	 * 		The exchange message to transfer
	 * 
	 * @return the list of Transport Protocols
	 * @throws TransportException
	 */
	private List<String> getTransportProtocolNames(
			MessageExchangeImpl exchange, String containerName)
			throws TransportException {
		log.start();

		List<String> listProtocols = null;

		String transportPolicy = (String) exchange
				.getProperty(Transporter.PROPERTY_TRANSPORT_QOS);
		
		// Get the message exchange default policy if null and set the property in the message
        if (MessageExchangeImpl.ROBUST_IN_ONLY_PATTERN.toString().equals(exchange.getPattern().toString())) {
            transportPolicy = Transporter.RELIABLE_POLICY;
            exchange.setProperty(Transporter.PROPERTY_TRANSPORT_QOS, transportPolicy);
        }

		// Get the system default policy if null and set the property in the message
		if (transportPolicy == null) {
			transportPolicy = SystemUtil.getTransporterQOS();
			exchange.setProperty(Transporter.PROPERTY_TRANSPORT_QOS,
					transportPolicy);
		}

		if (transportPolicy.equals(Transporter.FAST_POLICY)) {
			if (containerName.equals(containerInformation.getName())) {
				listProtocols = PROTOCOL_POLICIES_MAP.get(FAST_LOCAL_POLICY);
			} else {
				listProtocols = PROTOCOL_POLICIES_MAP.get(FAST_REMOTE_POLICY);
			}
		} else {
			listProtocols = PROTOCOL_POLICIES_MAP.get(transportPolicy);
		}

		if (listProtocols == null) {
			throw new TransportException(
					"The transport QOS specified in the message exchange is not valid: "
							+ transportPolicy);
		}

		log.end("Process the '" + transportPolicy + "' transport QOS");
		return listProtocols;
	}

	/**
	 * Utility to get the matching Transport Protocol component
	 * 
	 * @param protocolName
	 * 		The name of the Transport Protocol
	 */
	private TransportProtocol getProtocol(String protocolName)
			throws TransportException {

		TransportProtocol protocol = (TransportProtocol) protocols
				.get(protocolName);
		if (protocol == null)
			throw new TransportException(
					"The transport protocol linked to the interface '"
							+ protocolName + "' does not exist");

		return protocol;
	}
}
