/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: JoramTransporter.java,v 0.3 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.petals.jbi.messaging.transport.joram;

import java.io.File;
import java.io.IOException;
import java.util.Hashtable;
import java.util.List;
import java.util.concurrent.ArrayBlockingQueue;

import javax.jbi.messaging.MessageExchange;
import javax.jms.ExceptionListener;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.naming.InitialContext;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.petals.PetalsException;
import org.objectweb.petals.communication.jndi.JNDIService;
import org.objectweb.petals.communication.network.ContainerInformation;
import org.objectweb.petals.communication.network.NetworkService;
import org.objectweb.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.objectweb.petals.jbi.messaging.transport.TransportException;
import org.objectweb.petals.jbi.messaging.transport.TransportProtocol;
import org.objectweb.petals.jbi.messaging.transport.TransportProtocolListener;
import org.objectweb.petals.jbi.messaging.transport.Transporter;
import org.objectweb.petals.jbi.messaging.transport.util.JoramConfigurationBuilder;
import org.objectweb.petals.jbi.messaging.transport.util.JoramControl;
import org.objectweb.petals.jbi.messaging.transport.util.TransporterUtil;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.SystemUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * Implementation of the <code>TransporterProtocol</code> with JORAM.<br>
 * In dynamic mode, at the starting of the Transporter, the JORAM configuration
 * file is created according to the PEtALS topology, then the JORAM server is
 * started. For each component installed, the <code>JoramTransporter</code>
 * creates and registers a <code>Queue</code> in the distributed JNDI server
 * with the following name:
 * <em>JoramConnection-[containerName]-[componentName]</em><br>
 * The name of a container is the <em>ID</em> specified in the
 * <em>conf/server.properties</em> file. <br>
 * Similary, at each uninstallation of component, the <code>Queue</code> is
 * deleted and unregistered. <br>
 * This <code>JoramTransporter</code> uses an <code>ObjectSerializer</code>
 * to create <em>jms.ObjectMessage</em> to send via JMS. So, the
 * <code>MessageExchange</code> to send is serialized.
 *
 * @version $Rev: 477 $ $Date: 2006-05-29 17:18:07 +0200 (lun., 29 mai 2006) $
 * @since Petals 1.0
 * @author alouis
 * @author rnaudin
 * @author garciap
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = TransportProtocol.class))
public class JoramTransporter implements TransportProtocol {

    /**
     * The logging util
     */
    protected LoggingUtil log;

    /**
     * The Joram Connection instance
     */
    private JoramConnection joramConnection;

    /**
     * The local container information.
     */
    protected ContainerInformation containerInformation;

    /**
     * Joram configuration builder singleton.
     */
    private JoramConfigurationBuilder configurationBuilder = JoramConfigurationBuilder
            .getInstance();

    /**
     * Joram controller singleton.
     */
    private JoramControl joramControl = JoramControl.getInstance();

    /**
     * Map of blocking queues for the received exchanges waiting to be consumed
     * by the dispatcher.
     */
    private Hashtable<String, ArrayBlockingQueue<MessageExchange>> asyncExchangesQueuesMap;
    
    /**
     * The flag indicating the network time synchronization
     */
    private boolean timeSynchronized;

    /**
     * The Fractal component
     */
    @Service
    Component joramTransporterComponent;

    /**
     * Monolog component logger
     */
    @Monolog(name = "logger")
    protected Logger logger;

    @Requires(name = "network", signature = org.objectweb.petals.communication.network.NetworkService.class)
    protected NetworkService containerNetwork;

    @Requires(name = "jndi", signature = org.objectweb.petals.communication.jndi.JNDIService.class)
    protected JNDIService jndiService;

    @Requires(name = "dispatcher", signature = org.objectweb.petals.jbi.messaging.transport.TransportProtocolListener.class)
    protected TransportProtocolListener dispatcher;

    /**
     * Send the message to the specified distant or local container, by
     * serializing the exchange with a {@link ObjectSerializer}. We retrieve
     * the targeted {@link javax.jms.Queue} with the name of the container and
     * the component, transform the {@link MessageExchangeImpl} into a
     * {@link javax.jms.Message}, and send this message to the related
     * {@link javax.jms.Queue}. <br/> <br/> If the
     * "org.objectweb.petals.transport.nonpersistent" exchange property is set
     * to true, the message is not persisted by the JORAM layer, but it can be
     * lost in case of problem.
     *
     * @see org.objectweb.petals.jbi.messaging.transport.Transporter#send(org.objectweb.petals.jbi.messaging.exchange.MessageExchangeImpl,
     *      java.lang.String, java.lang.String, long)
     */

    public void send(MessageExchangeImpl messageExchange, final String componentName,
            final String containerName, final long timeToLive) throws TransportException {

        if (messageExchange == null) {
            throw new NullPointerException("Parameter messageExchange must not be null");
        } else if (containerName == null) {
            throw new NullPointerException("Parameter containerName must not be null");
        } else if (componentName == null) {
            throw new NullPointerException("Parameter componentName must not be null");
        } else if (timeToLive < 0) {
            throw new NullPointerException("Parameter timeToLive must be positive or zero");
        }

        log.start("Send message to container : " + containerName + " / component : "
                + componentName);

        boolean persistent = true;
        if (messageExchange.getProperty(Transporter.PROPERTY_TRANSPORT_QOS) != null
                && Transporter.FAST_POLICY.equals((String) messageExchange
                        .getProperty(Transporter.PROPERTY_TRANSPORT_QOS))) {
            persistent = false;
        }
        
        try {
            joramConnection.sendToDestination(
                    containerName, componentName, messageExchange,
                    // if the network is not time synchronized, don't set any
                    // TTL on the
                    // message transfer
                    !timeSynchronized ? Message.DEFAULT_TIME_TO_LIVE : timeToLive,
                    TransporterUtil.getSyncMode(messageExchange, true),
                    persistent);
            
        } catch (TransportException e) {
            log.error(e.getMessage(), e);
            throw e;
        }
        log.end();
    }

    /**
     * Initialize the fields. In dynamic mode, build the JORAM topology, then
     * start the JORAM server.
     *
     * @throws Exception
     */
    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws Exception {
        log = new LoggingUtil(logger);
        log.start();

        asyncExchangesQueuesMap = new Hashtable<String, ArrayBlockingQueue<MessageExchange>>();

        // get the local information
        containerInformation = containerNetwork.getLocalContainerInformation();
        
        timeSynchronized = Boolean.parseBoolean(SystemUtil.getSynchronizedTime().trim());

        // build the JORAM topology configuration file in dynamic mode
        if (SystemUtil.isDynamicTopology()) {
            doDynamicConfiguration();
        }
        
        // if JORAM is not launched (by JNDI components), initialize and start it
        if (!joramControl.isAgentServerRunning()) {
            // initialize the JORAM agent
            joramControl.initAgentServer(new File(SystemUtil.getJoramPath()),
                    Short.parseShort(containerInformation.getJoramId()));

            // start the JORAM agent
            try {
                joramControl.startAgentServer();
            } catch (Exception e) {
                throw new PetalsException("Can not start JORAM agent", e);
            }
        }

        // get the local container from the JNDI registry
        InitialContext rootContext = jndiService.getInitialContext();
        if (rootContext == null) {
            throw new NullPointerException("rootContext must not be null");
        }

        TransportProtocol joramTransportProtocol = (TransportProtocol) joramTransporterComponent
                .getFcInterface("service");

        joramConnection = new JoramConnection(joramTransportProtocol, containerInformation,
                rootContext, log);

        // create an exception listener to log all asynchronious errors
        ExceptionListener exceptionListener = new ExceptionListener() {
            public void onException(JMSException exception) {
                log.error("Asynchronious JMS exception", exception);
                return;
            }
        };

        try {
            joramConnection.start(exceptionListener);
        } catch (Exception e) {
            throw new TransportException(e);
        }

        log.end();
    }

    /**
     * Close the JMS connection and stop the JORAM server. The
     * <code>Transporter</code> stay registered in the distributed JNDI
     * server.
     *
     * @throws TransportException
     */
    @LifeCycle(on = LifeCycleType.STOP)
    public void stop() throws TransportException {
        log.start();

        try {
            joramConnection.stop();

            if (SystemUtil.isDynamicTopology()) {
                joramControl.stopAgentServer();
            }
        } catch (Exception e) {
            String msg = "Can not stop the Joram Transporter";
            log.error(msg, e);
            throw new TransportException(msg, e);
        }

        log.end();
        logger = null;
    }

    /**
     * Shutdown the <code>Transporter</code> and unregister it from the
     * network. The embedded component's Topics must have been removed from
     * Joram and from the distributed Jndi server previously.
     *
     * @throws TransportException
     */
    public void shutdown() throws TransportException {
        log.start();

        try {
            if (SystemUtil.isDynamicTopology())
                // TODO: remove the unregister container
                containerNetwork.unregisterContainer(containerInformation.getName());
            // remove the joram server from the topology and stop the server
            if (containerNetwork.retrieveAllContainersInformation(null).size() > 0) {
                joramControl.removeContainer(containerInformation);
            }
        } catch (PetalsException e) {
            String msg = "Can not shutdown unregister correctly from the network";
            log.error(msg, e);
            throw new TransportException(msg, e);
        }
        log.end();
    }

    /**
     * The <code>Transporter</code> creates a topic for the given component
     * and bind it to the JNDI server. It creates a subscriber to poll the
     * 'normal' messages and a listener to receiving the response messages when
     * synchronious send has been done.
     *
     * @throws TransportException
     */
    public void addDestination(final String componentName) throws TransportException {
        if (componentName == null) {
            throw new NullPointerException("Parameter componentName must not be null");
        }
        log.start("Add destination : " + componentName);

        ArrayBlockingQueue<MessageExchange> asyncExchangesQueue = new ArrayBlockingQueue<MessageExchange>(
                JoramConnection.RECEIVE_POOLSIZE);

        asyncExchangesQueuesMap.put(componentName, asyncExchangesQueue);

        try {
            joramConnection.createQueueDevice(componentName, dispatcher, asyncExchangesQueue);
        } catch (Exception e) {
            String msg = "Cannot add the destination for the component '" + componentName + "'";
            log.error(msg, e);
            throw new TransportException(msg, e);
        }
        log.end();
    }

    /**
     * The <code>Transporter</code> removes the <code>Sesssions</code> bound
     * to the component related Topic. Then it deletes the Topic and unbind it
     * from the JNDI server.
     */
    public void removeDestination(final String componentName) throws TransportException {
        if (componentName == null) {
            throw new NullPointerException("Parameter componentName must not be null");
        }
        log.start("Remove destination : " + componentName);

        try {
            joramConnection.deleteQueueDevice(componentName);
        } catch (Exception e) {
            String msg = "Cannot remove the destination for the component '" + componentName + "'";
            log.error(msg, e);
            throw new TransportException(msg, e);
        }
        log.end();
    }

    /**
     * The <code>Transporter</code> polls for a received message on the
     * component related topic.
     */
    public MessageExchangeImpl receive(final String componentName) throws TransportException {
        if (componentName == null) {
            throw new NullPointerException("Parameter componentName must not be null");
        }
        log.start("Receive : " + componentName);

        ArrayBlockingQueue<MessageExchange> asyncExchangesQueue = asyncExchangesQueuesMap
                .get(componentName);

        MessageExchangeImpl exchange;
        exchange = (MessageExchangeImpl) asyncExchangesQueue.poll();

        // notify the exchange to awake the async message listener, which
        // acknowledge the JMS message.
        synchronized (exchange) {
            exchange.notify();
        }

        log.end();

        return exchange;
    }

    /**
     * TRIBe based platform. It retrieves the distributed configuration and
     * builds the JORAM configuration file.
     * 
     * @throws PetalsException
     */
    private void doDynamicConfiguration() throws PetalsException {

        // Build the configuration file for JORAM ...
        List<ContainerInformation> containers = containerNetwork
                .retrieveAllContainersInformation(null);
        String config = configurationBuilder.retrieveDistributedConfiguration(containerInformation,
                containers);

        // save configuration to file
        File joramDir = new File(SystemUtil.getJoramPath());
        try {
            configurationBuilder.writeConfiguration(config, joramDir);
        } catch (IOException e) {
            throw new PetalsException("Can not write JORAM configuration file to dir "
                    + joramDir.getAbsolutePath(), e);
        }
    }

}
