/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: PetalsAdminService.java 16:40:52 alouis $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.kernel.admin;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jbi.JBIException;
import javax.naming.NamingException;

import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.petals.PetalsException;
import org.objectweb.petals.communication.network.ContainerInformation;
import org.objectweb.petals.communication.network.NetworkService;
import org.objectweb.petals.jbi.management.deployment.DeploymentServiceMBean;
import org.objectweb.petals.jbi.management.installation.InstallationServiceMBean;
import org.objectweb.petals.jbi.messaging.registry.EndpointService;
import org.objectweb.petals.jbi.messaging.servicedesc.AbstractEndpoint;
import org.objectweb.petals.jbi.messaging.transport.Transporter;
import org.objectweb.petals.kernel.server.PetalsServer;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.SystemUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * This class is used for Petals global Administration
 * 
 * @author alouis - eBMWebsourcing
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.objectweb.petals.kernel.admin.PetalsAdminServiceInterface.class))
public class PetalsAdminService implements PetalsAdminServiceInterface {

    /**
     * JBI :: JMX Deployment
     */
    @Requires(name = "deployment", signature = org.objectweb.petals.jbi.management.deployment.DeploymentServiceMBean.class)
    protected DeploymentServiceMBean deployment;

    /**
     * JBI :: JMX Installation
     */
    @Requires(name = "installation", signature = org.objectweb.petals.jbi.management.installation.InstallationServiceMBean.class)
    protected InstallationServiceMBean installation;

    /**
     * Logger wrapper
     */
    protected LoggingUtil log;

    /**
     * Monolog
     */
    @Monolog(name = "logger")
    protected Logger logger;

    @Requires(name = "endpoint", signature = org.objectweb.petals.jbi.messaging.registry.EndpointService.class)
    protected EndpointService endpoint;

    @Requires(name = "networkService", signature = org.objectweb.petals.communication.network.NetworkService.class)
    protected NetworkService networkService;

    @Requires(name = "transporter", signature = org.objectweb.petals.jbi.messaging.transport.Transporter.class)
    protected Transporter transporter;

    /**
     * The Petals Server instance, which manages the Fractal composite
     */
    private PetalsServer petalsServer;

    /**
     * Retrieve the collection of all the started servers configuration from
     * this domain.
     * 
     * @return Collection of server information ContainerInformation
     */
    public Collection<Map<String, String>> getAllServersConfiguration() {
        return retrieveAllServersConfiguration(false);
    }

    /**
     * Retrieve the collection of all the started servers configuration from
     * this domain.
     * 
     * @return Collection of server information ContainerInformation
     */
    public Collection<Map<String, String>> getAllStartedServersConfiguration() {
        return retrieveAllServersConfiguration(true);
    }

    /**
     * Remove the specified server from the network configuration. <br>
     * The server is unbound, but not shutdown. All ServiceEndpoints activated
     * on this server are unbound. <br>
     * <b>Be very careful when using this operation, the network configuration
     * may become unstable.</b><br>
     * TODO JORAM configuration not updated
     * 
     * @param serverName
     * @throws Exception
     *      The server is not found in the network configuration, some Endpoints
     *      have not been removed, or problem while accessing the network
     *      configuration
     */
    public void removeServerFromNetwork(String serverName) throws Exception {

        log.start();

        try {
            // Unregister container endpoints
            final Collection<AbstractEndpoint> eps = this.endpoint.getServiceEndpointForContainer(serverName);
            for (final AbstractEndpoint endpt : eps) {
                this.endpoint.deactivateEndpoint(endpt);
            }
            
            // Unregister container
            this.networkService.unregisterContainer(serverName);

        } catch (final PetalsException e) {
            this.log.error("Error when removing a server.", e);
            throw new Exception("Error when removing a server.", e);
        } catch (NamingException e) {
            this.log.error("Error when retrieving container endpoints.", e);
            throw new Exception("Error when retrieving container endpoints.", e);
        } catch (final JBIException e) {
            this.log.error("A ServiceEndpoint failed to be removed.", e);
            throw new Exception("A ServiceEndpoint failed to be removed.", e);
        }

        log.end();
    }

    public void setPetalsServer(PetalsServer petalsServer) {
        this.petalsServer = petalsServer;
    }

    /**
     * Uninstalled serviceassemblies, components and sharelibraries endpoints
     * are removed the container is removed from the global configuration, then
     * stop the Petals container
     * 
     * @throws Exception
     */
    public void shutdownContainer() throws Exception {
        log.start();

        log.info("Shutdown the server. It will be removed from the Petals network.");
        try {
            log.info("  1/4 Undeploy Service Assemblies.");
            deployment.shutdown();
        } catch (Exception e) {
            log.error("Can not shutdown the DeploymentServiceMBean.", e);
        }
        try {
            log.info("  2/4 Uninstall Shared Libraries and Components.");
            installation.shutdown();
        } catch (Exception e) {
            log.error("CInstallationServiceMBean InstallationService.", e);
        }
        if (SystemUtil.isDynamicTopology()) {
            try {
                log
                        .info("  3/4 Shutdown the Transporter (remove the server from the global configuration).");
                transporter.shutdown();
            } catch (Exception e) {
                log.error("Can not shutdown the Transporter.", e);
            }
        }

        if (SystemUtil.isDynamicTopology()) {
            log.info("  4/4 Stopping the server...");
        } else {
            log.info("  3/4 Stopping the server...");
        }
        // launch the stop of the Petals server in a different thread
        petalsServer.getPetalsStopThread().start();
        log.end();
    }

    /**
     * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
     */
    @LifeCycle(on = LifeCycleType.START)
    public void start() throws IllegalLifeCycleException {
        log = new LoggingUtil(logger);
    }

    // ----------------------------------------------------------------------
    // Fractal LifecycleController implementation
    // ----------------------------------------------------------------------

    /**
     * Stop the Petals container
     * 
     * @throws Exception
     */
    public void stopContainer() throws Exception {
        log.start();
        log.info("Stopping the server...");
        // launch the stop of the Petals server in a different thread
        petalsServer.getPetalsStopThread().start();
        log.end();
    }

    /**
     * Create a Map from a ContainerInformation object
     * 
     * @param info
     * @return a Map of the server properties
     */
    private static Map<String, String> createConfigurationMap(ContainerInformation info) {
        Map<String, String> map = new HashMap<String, String>();
        if (info != null) {
            map.put(CONF_HOST, info.getHost());
            map.put(CONF_HTML_PORT, info.getHtmlPort());
            map.put(CONF_JMX_LOGIN, info.getJmxLogin());
            map.put(CONF_JMX_PASSWORD, info.getJmxPassword());
            map.put(CONF_JMX_PORT, info.getJmxPort());
            map.put(CONF_JMX_JNDI_PORT, info.getJmxJndiPort());
            map.put(CONF_JNDI_FACTORY, info.getJndiFactory());
            map.put(CONF_JNDI_PORT, info.getJndiPort());
            map.put(CONF_JORAM_DOMAIN, info.getJoramDomain());
            map.put(CONF_JORAM_DOMAIN_PORT, info.getJoramDomainPort());
            map.put(CONF_JORAM_ID, info.getJoramId());
            map.put(CONF_JORAM_LOGIN, info.getJoramLogin());
            map.put(CONF_JORAM_PASSWORD, info.getJoramPassword());
            map.put(CONF_JORAM_TCP_PORT, info.getJoramTCPPort());
            map.put(CONF_NAME, info.getName());
            map.put(CONF_STATE,
                    (info.getState() == ContainerInformation.ContainerState.STARTED) ? "started"
                            : "stopped");
            map.put(CONF_UID, "" + info.getUid());
        }
        return map;
    }

    /**
     * Retrieve the collection of all the started servers configuration from
     * this domain.
     * 
     * @param startedOnly
     *      return only started servers information
     * @return Collection of server information ContainerInformation
     */
    protected Collection<Map<String, String>> retrieveAllServersConfiguration(boolean startedOnly) {

        log.start();
        
        final Set<Map<String, String>> containers = new HashSet<Map<String, String>>();

        final List<ContainerInformation> containersInformation = this.networkService.retrieveAllContainersInformation(null);
        
        for (final ContainerInformation containerInformation : containersInformation) {
            if (!startedOnly || (startedOnly && containerInformation.isStarted())) {

                containers.add(PetalsAdminService.createConfigurationMap(containerInformation));
            }
        }

        log.end();
        return containers;
    }

}
