/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2009 OW2, http://www.ow2.org/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: $
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.extensions.transporter.nio.server;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.petals.jbi.messaging.transport.TransportProtocol;
import org.objectweb.petals.jbi.messaging.transport.Transporter;
import org.objectweb.petals.kernel.configuration.LocalConfiguration;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.extensions.transporter.nio.Parameters;
import org.ow2.petals.extensions.transporter.nio.selector.InputSocketChannelContext;
import org.ow2.petals.extensions.transporter.nio.selector.NioSelector;
import org.ow2.petals.extensions.transporter.nio.selector.NioServer;
import org.ow2.petals.extensions.transporter.nio.selector.PipedDeserializer;

/**
 * The server agent of the NIO transporter.
 * 
 * @author Christophe DENEUX (Capgemini Sud)
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = NioServer.class))
public class NioServerAgent implements NioServer {

    /**
     * <p>
     * The socket selector Fractal component to use to process network IO.
     * </p>
     */
    @Requires(name = NioSelector.NIO_SELECTOR_ITF, signature = NioSelector.class)
    protected NioSelector selectorAgent;

    @Requires(name = NioReceiver.NIO_MSG_RECEIVER_ITF, signature = NioReceiver.class)
    protected NioReceiver messageReceiver;

    /**
     * Interface used to access configuration
     */
    @Requires(name = TransportProtocol.CONFIGURATION_ITF, signature = LocalConfiguration.class)
    protected LocalConfiguration configurationService;

    /**
     * Fractal Logger
     */
    @Monolog(name = "logger")
    protected Logger logger;

    private LoggingUtil log;

    /**
     * The server socket on which connections are waited.
     */
    private InetSocketAddress serverSocket;

    /**
     * The server socket channel on which connections are waited.
     */
    private ServerSocketChannel serverSocketChannel;

    /**
     * Map containing the deserialization task associated to each socket
     * channel.
     */
    private final Map<SocketChannel, PipedDeserializer> deserializationTasks = new ConcurrentHashMap<SocketChannel, PipedDeserializer>();

    /**
     * Acceptor thread pool. An acceptor is in charge to deserialize received
     * data into its associated object and transmit it to the
     * {@link Transporter}.
     */
    private ThreadPoolExecutor threadPoolTaskExecutor = null;

    @LifeCycle(on = LifeCycleType.START)
    public void start() throws IOException {

        this.log = new LoggingUtil(this.logger);
        this.log.start();

        final int listenPort = Integer.parseInt(this.configurationService
				.getValue(Parameters.PROP_NAME_LISTEN_PORT,
						Parameters.DEFAULT_LISTEN_PORT));
        this.log.debug("Starting NIO transporter server on port " + listenPort);
        this.serverSocket = new InetSocketAddress(listenPort);
        this.serverSocketChannel = ServerSocketChannel.open();

        // Caution, as the queue of the thread pool is a LinkedBlockingQueue,
        // the parameters maxPoolSize and keep alive time have no sense.
        // TODO: When using JDK6, we can use
        // ThreadPoolExecutor.allowCoreThreadTimeOut to decrease the number of
        // thread in the pool.
        final int nbAcceptors = Integer.parseInt(this.configurationService.getValue(
        		Parameters.PROP_NAME_ACCEPTORS, Parameters.DEFAULT_ACCEPTORS));
        this.threadPoolTaskExecutor = new ThreadPoolExecutor(nbAcceptors, nbAcceptors, 5,
                TimeUnit.SECONDS, new LinkedBlockingQueue<Runnable>(),
                new DeserializerThreadFactory());

        // TODO: Set socket options
        // this.serverSocketChannel.socket().setPerformancePreferences(0, 1, 2);

        this.serverSocketChannel.socket().bind(this.serverSocket, 100);
        this.serverSocketChannel.configureBlocking(false);

        this.log.info("The NIO transporter server is ready to process request on port "
                + listenPort);

        this.selectorAgent.register(SelectionKey.OP_ACCEPT, this.serverSocketChannel, this);
    }

    @LifeCycle(on = LifeCycleType.STOP)
    public void stop() throws Exception {
        this.log.start();

        this.log.debug("Stopping NIO Server agent.");
        this.serverSocketChannel.close();
        this.log.info("The NIO Server agent is stopped.");

        this.threadPoolTaskExecutor.shutdownNow();

        this.log.end();
    }

    public void onConnectionIsAccepted(final SocketChannel socketChannel, final Selector selector)
            throws IOException {

        this.log.info("A connection is accepted: " + socketChannel.socket().toString());

        final PipedDeserializer deserializer = new PipedDeserializerImpl(socketChannel,
                this.messageReceiver, this.logger, this.log);
        this.deserializationTasks.put(socketChannel, deserializer);

        socketChannel.register(selector, SelectionKey.OP_READ, new InputSocketChannelContext(
                socketChannel, deserializer, this));
    }

    public void onRead(final PipedDeserializer deserializer) {

        try {
            this.threadPoolTaskExecutor.execute(deserializer);
        } catch (final RejectedExecutionException e) {
            this.log.error("A message deserialization task has been rejected (" + e.getMessage()
                    + ")");
        }
    }

    public void onClose(final SocketChannel socketChannel) {
        // We free the deserializer associated to the socket channel

        this.deserializationTasks.remove(socketChannel);

    }
}
